import copy
import logging
import os
import pickle

import numpy as np
from scipy.interpolate import CubicSpline

# Keras Imports
try:
    import keras
    from keras.models import Sequential, clone_model
    from keras.layers import Dense, Activation, Dropout
    from keras.optimizers import SGD
except ImportError:
    raise ImportError(
        "Could not import the required Keras modules. The QRNN "
        "implementation was developed for use with Keras version 2.0.9.")

################################################################################
# Loss Functions
################################################################################

import keras.backend as K


logger = logging.getLogger(__name__)


def skewed_absolute_error(y_true, y_pred, tau):
    """
    The quantile loss function for a given quantile tau:

    L(y_true, y_pred) = (tau - I(y_pred < y_true)) * (y_pred - y_true)

    Where I is the indicator function.
    """
    dy = y_pred - y_true
    return K.mean((1.0 - tau) * K.relu(dy) + tau * K.relu(-dy), axis=-1)


def quantile_loss(y_true, y_pred, taus):
    """
    The quantiles loss for a list of quantiles. Sums up the error contribution
    from the each of the quantile loss functions.
    """
    e = skewed_absolute_error(
        K.flatten(y_true), K.flatten(y_pred[:, 0]), taus[0])
    for i, tau in enumerate(taus[1:]):
        e += skewed_absolute_error(K.flatten(y_true),
                                   K.flatten(y_pred[:, i + 1]),
                                   tau)
    return e


class QuantileLoss:
    """
    Wrapper class for the quantile error loss function. A class is used here
    to allow the implementation of a custom `__repr` function, so that the
    loss function object can be easily loaded using `keras.model.load`.

    Attributes:

        quantiles: List of quantiles that should be estimated with
                   this loss function.

    """

    def __init__(self, quantiles):
        self.__name__ = "QuantileLoss"
        self.quantiles = quantiles

    def __call__(self, y_true, y_pred):
        return quantile_loss(y_true, y_pred, self.quantiles)

    def __repr__(self):
        return "QuantileLoss(" + repr(self.quantiles) + ")"

################################################################################
# Keras Interface Classes
################################################################################


class TrainingGenerator:
    """
    This Keras sample generator takes the noise-free training data
    and adds independent Gaussian noise to each of the components
    of the input.

    Attributes:

        x_train: The training input, i.e. the brightness temperatures
                 measured by the satellite.
        y_train: The training output, i.e. the value of the retrieval
                 quantity.
        x_mean: A vector containing the mean of each input component.
        x_sigma: A vector containing the standard deviation of each
                 component.
        batch_size: The size of a training batch.
    """

    def __init__(self, x_train, x_mean, x_sigma, y_train, sigma_noise, batch_size):
        self.bs = batch_size

        self.x_train = x_train
        self.x_mean = x_mean
        self.x_sigma = x_sigma
        self.y_train = y_train
        self.sigma_noise = sigma_noise

        self.indices = np.random.permutation(x_train.shape[0])
        self.i = 0

    def __iter__(self):
        logger.info("iter...")
        return self

    def __next__(self):
        inds = self.indices[np.arange(self.i * self.bs,
                                      (self.i + 1) * self.bs)
                            % self.indices.size]
        x_batch = np.copy(self.x_train[inds, :])
        if not self.sigma_noise is None:
            x_batch += np.random.randn(*x_batch.shape) * self.sigma_noise
        x_batch = (x_batch - self.x_mean) / self.x_sigma
        y_batch = self.y_train[inds]

        self.i = self.i + 1

        # Shuffle training set after each epoch.
        if self.i % (self.x_train.shape[0] // self.bs) == 0:
            self.indices = np.random.permutation(self.x_train.shape[0])

        return (x_batch, y_batch)

# TODO: Make y-noise argument optional


class AdversarialTrainingGenerator:
    """
    This Keras sample generator takes the noise-free training data
    and adds independent Gaussian noise to each of the components
    of the input.

    Attributes:

        x_train: The training input, i.e. the brightness temperatures
                 measured by the satellite.
        y_train: The training output, i.e. the value of the retrieval
                 quantity.
        x_mean: A vector containing the mean of each input component.
        x_sigma: A vector containing the standard deviation of each
                 component.
        batch_size: The size of a training batch.
    """

    def __init__(self,
                 x_train,
                 x_mean,
                 x_sigma,
                 y_train,
                 sigma_noise,
                 batch_size,
                 input_gradients,
                 eps):
        self.bs = batch_size

        self.x_train = x_train
        self.x_mean = x_mean
        self.x_sigma = x_sigma
        self.y_train = y_train
        self.sigma_noise = sigma_noise

        self.indices = np.random.permutation(x_train.shape[0])
        self.i = 0

        # compile gradient function
        bs2 = self.bs // 2

        self.input_gradients = input_gradients
        self.eps = eps

    def __iter__(self):
        logger.info("iter...")
        return self

    def __next__(self):

        if self.i == 0:
            inds = np.random.randint(0, self.x_train.shape[0], self.bs)

            x_batch = np.copy(self.x_train[inds, :])
            if (self.sigma_noise):
                x_batch += np.random.randn(*x_batch.shape) * self.sigma_noise

            x_batch = (x_batch - self.x_mean) / self.x_sigma
            y_batch = self.y_train[inds]

        else:

            bs2 = self.bs // 2
            inds = np.random.randint(0, self.x_train.shape[0], bs2)

            x_batch = np.zeros((self.bs, self.x_train.shape[1]))
            y_batch = np.zeros((self.bs, 1))

            x_batch[:bs2, :] = np.copy(self.x_train[inds, :])
            if (self.sigma_noise):
                x_batch[:bs2, :] += np.random.randn(bs2, self.x_train.shape[1]) \
                    * self.sigma_noise
            x_batch[:bs2, :] = (x_batch[:bs2, :] - self.x_mean) / self.x_sigma
            y_batch[:bs2, :] = self.y_train[inds].reshape(-1, 1)
            x_batch[bs2:, :] = x_batch[:bs2, :]
            y_batch[bs2:, :] = y_batch[:bs2, :]

            if (self.i > 10):
                grads = self.input_gradients(
                    [x_batch[:bs2, :], y_batch[:bs2, :], [1.0]])[0]
                x_batch[bs2:, :] += self.eps * np.sign(grads)

        self.i = self.i + 1
        return (x_batch, y_batch)


# TODO: Make y-noise argument optional
class ValidationGenerator:
    """
    This Keras sample generator is similar to the training generator
    only that it returns the whole validation set and doesn't perform
    any randomization.

    Attributes:

        x_val: The validation input, i.e. the brightness temperatures
                 measured by the satellite.
        y_val: The validation output, i.e. the value of the retrieval
                 quantity.
        x_mean: A vector containing the mean of each input component.
        x_sigma: A vector containing the standard deviation of each
                 component.
    """

    def __init__(self, x_val, x_mean, x_sigma, y_val, sigma_noise):
        self.x_val = x_val
        self.x_mean = x_mean
        self.x_sigma = x_sigma

        self.y_val = y_val

        self.sigma_noise = sigma_noise

    def __iter__(self):
        return self

    def __next__(self):
        x_val = np.copy(self.x_val)
        if not self.sigma_noise is None:
            x_val += np.random.randn(*self.x_val.shape) * self.sigma_noise
        x_val = (x_val - self.x_mean) / self.x_sigma
        return (x_val, self.y_val)


class LRDecay(keras.callbacks.Callback):
    """
    The LRDecay class implements the Keras callback interface and reduces
    the learning rate according to validation loss reduction.

    Attributes:

        lr_decay: The factor c > 1.0 by which the learning rate is
                  reduced.
        lr_minimum: The training is stopped when this learning rate
                    is reached.
        convergence_steps: The number of epochs without validation loss
                           reduction required to reduce the learning rate.

    """

    def __init__(self, model, lr_decay, lr_minimum, convergence_steps):
        self.model = model
        self.lr_decay = lr_decay
        self.lr_minimum = lr_minimum
        self.convergence_steps = convergence_steps
        self.steps = 0

    def on_train_begin(self, logs={}):
        self.losses = []
        self.steps = 0
        self.min_loss = 1e30

    def on_epoch_end(self, epoch, logs={}):
        self.losses += [logs.get('val_loss')]
        if not self.losses[-1] < self.min_loss:
            self.steps = self.steps + 1
        else:
            self.steps = 0
        if self.steps > self.convergence_steps:
            lr = keras.backend.get_value(self.model.optimizer.lr)
            keras.backend.set_value(
                self.model.optimizer.lr, lr / self.lr_decay)
            self.steps = 0
            logger.info("\n Reduced learning rate to " + str(lr))

            if lr < self.lr_minimum:
                self.model.stop_training = True

        self.min_loss = min(self.min_loss, self.losses[-1])

################################################################################
# QRNN
################################################################################


class QRNN:
    r"""
    Quantile Regression Neural Network (QRNN)

    This class implements quantile regression neural networks and can be used
    to estimate quantiles of the posterior distribution of remote sensing
    retrievals.

    Internally, the QRNN uses a feed-forward neural network that is trained
    to minimize the quantile loss function

    .. math::
            \mathcal{L}_\tau(y_\tau, y_{true}) =
            \begin{cases} (1 - \tau)|y_\tau - y_{true}| & \text{ if } y_\tau < y_\text{true} \\
            \tau |y_\tau - y_\text{true}| & \text{ otherwise, }\end{cases}

    where :math:`x_\text{true}` is the expected value of the retrieval quantity
    and and :math:`x_\tau` is the predicted quantile. The neural network
    has one output neuron for each quantile to estimate.

    For the training, this implementation provides custom data generators that
    can be used to add Gaussian noise to the training data as well as adversarial
    training using the fast gradient sign method.

    This implementation also provides functionality to use an ensemble of networks
    instead of just a single network to predict the quantiles.

    .. note:: For the QRNN I am using :math:`x` to denote the input vector and
              :math:`y` to denote the output. While this is opposed to typical
              inverse problem notation, it is inline with machine learning
              notation and felt more natural for the implementation. If this
              annoys you, I am sorry. But the other way round it would have
              annoyed other people and in particular me.

    Attributes:

        input_dim (int):
            The input dimension of the neural network, i.e. the dimension of the
            measurement vector.

        quantiles (numpy.array):
            The 1D-array containing the quantiles :math:`\tau \in [0, 1]` that the
            network learns to predict.

        depth (int):
            The number layers in the network excluding the input layer.

        width (int):
            The width of the hidden layers in the network.

        activation (str):
            The name of the activation functions to use in the hidden layers
            of the network.

        models (list of keras.models.Sequential):
            The ensemble of Keras neural networks used for the quantile regression
            neural network.
    """

    def __init__(self,
                 input_dim,
                 quantiles,
                 depth=3,
                 width=128,
                 activation="relu",
                 ensemble_size=1,
                 **kwargs):
        """
        Create a QRNN model.

        Arguments:

            input_dim(int): The dimension of the measurement space, i.e. the number
                            of elements in a single measurement vector y

            quantiles(np.array): 1D-array containing the quantiles  to estimate of
                                 the posterior distribution. Given as fractions
                                 within the range [0, 1].

            depth(int): The number of hidden layers  in the neural network to
                        use for the regression. Default is 3, i.e. three hidden
                        plus input and output layer.

            width(int): The number of neurons in each hidden layer.

            activation(str): The name of the activation functions to use. Default
                             is "relu", for rectified linear unit. See 
                             `this <https://keras.io/activations>`_ link for
                             available functions.

            **kwargs: Additional keyword arguments are passed to the constructor
                      call `keras.layers.Dense` of the hidden layers, which can
                      for example be used to add regularization. For more info consult
                      `Keras documentation. <https://keras.io/layers/core/#dense>`_
        """
        self.input_dim = input_dim
        self.quantiles = np.array(quantiles)
        self.depth = depth
        self.width = width
        self.activation = activation

        model = Sequential()
        if depth == 0:
            model.add(Dense(input_dim=input_dim,
                            units=len(quantiles),
                            activation=None))
        else:
            model.add(Dense(input_dim=input_dim,
                            units=width,
                            activation=activation))
            for i in range(depth - 2):
                model.add(Dense(units=width,
                                activation=activation,
                                **kwargs))
            model.add(Dense(units=len(quantiles), activation=None))
        self.models = [clone_model(model) for i in range(ensemble_size)]

    def __fit_params__(self, kwargs):
        at = kwargs.pop("adversarial_training", False)
        dat = kwargs.pop("delta_at", 0.01)
        batch_size = kwargs.pop("batch_size", 512)
        convergence_epochs = kwargs.pop("convergence_epochs", 10)
        initial_learning_rate = kwargs.pop('initial_learning_rate', 0.01)
        learning_rate_decay = kwargs.pop('learning_rate_decay', 2.0)
        learning_rate_minimum = kwargs.pop('learning_rate_minimum', 1e-6)
        maximum_epochs = kwargs.pop("maximum_epochs", 200)
        training_split = kwargs.pop("training_split", 0.9)
        return at, dat, batch_size, convergence_epochs, initial_learning_rate, \
            learning_rate_decay, learning_rate_minimum, maximum_epochs, \
            training_split, kwargs

    def cross_validation(self,
                        x_train,
                        y_train,
                        sigma_noise = None,
                        n_folds=5,
                        s=None,
                        **kwargs):
        r"""
        Perform n-fold cross validation.

        This function trains the network n times on different subsets of the
        provided training data, always keeping a fraction of 1/n samples apart
        for testing. Performance for each of the networks is evaluated and mean
        and standard deviation for all folds are returned. This is to reduce
        the influence of random fluctuations of the network performance during
        hyperparameter tuning.

        Arguments:

            x_train(numpy.array): Array of shape :code:`(m, n)` containing the
                                  m n-dimensional training inputs.

            y_train(numpy.array): Array of shape :code:`(m, 1)` containing the
                                  m training outputs.

            sigma_noise(None, float, np.array): If not `None` this value is used
                                                to multiply the Gaussian noise
                                                that is added to each training
                                                batch. If None no noise is
                                                added.

            n_folds(int): Number of folds to perform for the cross correlation.

            s(callable, None): Performance metric for the fold. If not None,
                               this should be a function object taking as
                               arguments :code:`(y_test, y_pred)`, i.e. the
                               expected output for the given fold :code:`y_test`
                               and the predicted output :code:`y_pred`. The
                               returned value is taken as performance metric.

            **kwargs: Additional keyword arguments are passed on to the :code:`fit`
                      method that is called for each fold.
        """

        n = x_train.shape[0]
        n_test = n // n_folds
        inds = np.random.permutation(np.arange(0, n))

        results = []

        # Nomenclature is a bit difficult here:
        # Each cross validation fold has its own training,
        # vaildation and test set. The size of the test set
        # is number of provided training samples divided by the
        # number of fold. The rest is used a traning and internal
        # validation set according to the chose training_split
        # ratio.


        for i in range(n_folds):
            for m in self.models:
                m.reset_states()

            # Indices to use for training including training data and internal
            # validation set to monitor convergence.
            inds_train = np.append(np.arange(0, i * n_test),
                                       np.arange(min((i + 1) * n_test, n), n))
            inds_train = inds[inds_train]
            # Indices used to evaluate performance of the model.
            inds_test = np.arange(i * n_test, (i + 1) * n_test)
            inds_test = inds[inds_test]

            x_test_fold = x_train[inds_test, :]
            y_test_fold = y_train[inds_test]

            # Splitting training and validation set.
            x_train_fold = x_train[inds_train, :]
            y_train_fold = y_train[inds_train]

            self.fit(x_train_fold, y_train_fold,
                     sigma_noise, **kwargs)

            # Evaluation on this folds test set.
            if s:
                y_pred = self.predict(x_test_fold)
                results += [s(y_pred, y_test_fold)]
            else:
                loss = self.models[0].evaluate(
                    (x_test_fold - self.x_mean) / self.x_sigma,
                    y_test_fold)
                logger.info(loss)
                results += [loss]
        logger.info(results)
        results = np.array(results)
        logger.info(results)
        return (np.mean(results, axis=0), np.std(results, axis=0))

    def fit(self,
            x_train,
            y_train,
            sigma_noise=None,
            x_val=None,
            y_val=None,
            **kwargs):
        r"""
        Train the QRNN on given training data.

        The training uses an internal validation set to monitor training
        progress. This can be either split at random from the training
        data (see `training_fraction` argument) or provided explicitly
        using the `x_val` and `y_val` parameters

        Training of the QRNN is performed using stochastic gradient descent
        (SGD). The learning rate is adaptively reduced during training when
        the loss on the internal validation set has not been reduced for a
        certain number of epochs.

        Two data augmentation techniques can be used to improve the
        calibration of the QRNNs predictions. The first one adds Gaussian
        noise to training batches before they are passed to the network.
        The noise statistics are defined using the `sigma_noise` argument.
        The second one is adversarial training. If adversarial training is
        used, half of each training batch consists of adversarial samples
        generated using the fast gradient sign method. The strength of the
        perturbation is controlled using the `delta_at` parameter.

        During one epoch, each training sample is passed once to the network.
        Their order is randomzied between epochs.

        Arguments:

            x_train(np.array): Array of shape `(n, m)` containing n training
                               samples of dimension m.

            y_train(np.array): Array of shape `(n, )` containing the training
                               output corresponding to the training data in
                               `x_train`.

            sigma_noise(None, float, np.array): If not `None` this value is used
                                                to multiply the Gaussian noise
                                                that is added to each training
                                                batch. If None no noise is
                                                added.
            x_val(np.array): Array of shape :code:`(n', m)` containing n' validation
                             inputs that will be used to monitor training loss. Must
                             be provided in unison with :code:`y_val` or otherwise
                             will be ignored.

            y_val(np.array): Array of shape :code:`(n')` containing n'  validation
                             outputs corresponding to the inputs in :code:`x_val`.
                             Must be provided in unison with :code:`x_val` or
                             otherwise will be ignored.

            adversarial_training(Bool): Whether or not to use adversarial training.
                                        `False` by default.

            delta_at(flaot): Perturbation factor for the fast gradient sign method
                             determining the strength of the adversarial training
                             perturbation. `0.01` by default.

            batch_size(float): The batch size to use during training. Defaults to `512`.

            convergence_epochs(int): The number of epochs without decrease in
                                     validation loss before the learning rate
                                     is reduced. Defaults to `10`.

            initial_learning_rate(float): The inital value for the learning
                                          rate.

            learning_rate_decay(float): The factor by which to reduce the
                                        learning rate after no improvement
                                        on the internal validation set was
                                        observed for `convergence_epochs`
                                        epochs. Defaults to `2.0`.

            learning_rate_minimum(float): The minimum learning rate at which
                                          the training is terminated. Defaults
                                          to `1e-6`.

            maximum_epochs(int): The maximum number of epochs to perform if
                                 training does not terminate before.

            training_split(float): The ratio `0 < ts < 1.0` of the samples in
                                   to be used as internal validation set. Defaults
                                   to 0.9.

        """
        if not (x_train.shape[1] == self.input_dim):
            raise Exception("Training input must have the same extent along"    /
                            "dimension 1 as input_dim (" + str(self.input_dim)  /
                            + ")")

        if not (y_train.shape[1] == 1):
            raise Exception("Currently only scalar retrieval targets are"       /
                            " supported.")

        x_mean = np.mean(x_train, axis=0, keepdims=True)
        x_sigma = np.std(x_train, axis=0, keepdims=True)
        self.x_mean = x_mean
        self.x_sigma = x_sigma

        # Handle parameters
        # at:  adversarial training
        # bs:  batch size
        # ce:  convergence epochs
        # ilr: initial learning rate
        # lrd: learning rate decay
        # lrm: learning rate minimum
        # me:  maximum number of epochs
        # ts:  split ratio of training set
        at, dat, bs, ce, ilr, lrd, lrm, me, ts, kwargs = self.__fit_params__(
            kwargs)

        # Split training and validation set if x_val or y_val
        # are not provided.
        n = x_train.shape[0]
        n_train = n
        if x_val is None and y_val is None:
            n_train = round(ts * n)
            n_val = n - n_train
            inds = np.random.permutation(n)
            x_val = x_train[inds[n_train:], :]
            y_val = y_train[inds[n_train:]]
            x_train = x_train[inds[:n_train], :]
            y_train = y_train[inds[:n_train]]
        loss = QuantileLoss(self.quantiles)

        self.custom_objects = {loss.__name__: loss}
        for model in self.models:
            optimizer = SGD(lr=ilr)
            model.compile(loss=loss, optimizer=optimizer)

            if at:
                inputs = [model.input, model.targets[0],
                          model.sample_weights[0]]
                input_gradients = K.function(
                    inputs, K.gradients(model.total_loss, model.input))
                training_generator = AdversarialTrainingGenerator(x_train,
                                                                  self.x_mean,
                                                                  self.x_sigma,
                                                                  y_train,
                                                                  sigma_noise,
                                                                  bs,
                                                                  input_gradients,
                                                                  dat)
            else:
                training_generator = TrainingGenerator(x_train, self.x_mean, self.x_sigma,
                                                       y_train, sigma_noise, bs)
            validation_generator = ValidationGenerator(x_val, self.x_mean, self.x_sigma,
                                                       y_val, sigma_noise)
            lr_callback = LRDecay(model, lrd, lrm, ce)
            model.fit_generator(training_generator, steps_per_epoch=n_train // bs,
                                epochs=me, validation_data=validation_generator,
                                validation_steps=1, callbacks=[lr_callback])

    def predict(self, x):
        r"""
        Predict quantiles of the conditional distribution P(y|x).

        Forward propagates the inputs in `x` through the network to
        obtain the predicted quantiles `y`.

        Arguments:

            x(np.array): Array of shape `(n, m)` containing `n` m-dimensional inputs
                         for which to predict the conditional quantiles.

        Returns:

             Array of shape `(n, k)` with the columns corresponding to the k
             quantiles of the network.

        """
        predictions = np.stack(
            [m.predict((x - self.x_mean) / self.x_sigma) for m in self.models])
        return np.mean(predictions, axis=0)

    def cdf(self, x):
        r"""
        Approximate the posterior CDF for given inputs `x`.

        Propagates the inputs in `x` forward through the network and
        approximates the posterior CDF by a piecewise linear function.

        The piecewise linear function is given by its values at
        approximate quantiles $x_\tau$ for
        :math: `\tau = \{0.0, \tau_1, \ldots, \tau_k, 1.0\}` where
        :math: `\tau_k` are the quantiles to be estimated by the network.
        The values for :math:`x_0.0` and :math:`x_1.0` are computed using

        .. math::
            x_0.0 = 2.0 x_{\tau_1} - x_{\tau_2}

            x_1.0 = 2.0 x_{\tau_k} - x_{\tau_{k-1}}

        Arguments:

            x(np.array): Array of shape `(n, m)` containing `n` inputs for which
                         to predict the conditional quantiles.

        Returns:

            Tuple (xs, fs) containing the :math: `x`-values in `xs` and corresponding
            values of the posterior CDF :math: `F(x)` in `fs`.

        """
        y_pred = np.zeros(self.quantiles.size + 2)
        y_pred[1:-1] = self.predict(x)
        y_pred[0] = 2.0 * y_pred[1] - y_pred[2]
        y_pred[-1] = 2.0 * y_pred[-2] - y_pred[-3]

        qs = np.zeros(self.quantiles.size + 2)
        qs[1:-1] = self.quantiles
        qs[0] = 0.0
        qs[-1] = 1.0

        return y_pred, qs

    def pdf(self, x, use_splines = False):
        r"""
        Approximate the posterior probability density function (PDF) for given
        inputs `x`.

        By default, the PDF is approximated by computing the derivative of the
        piece-wise linear approximation of the CDF as computed by the :code:`cdf`
        function.

        If :code:`use_splines` is set to :code:`True`, the PDF is computed from
        a spline fit to the approximate CDF.

        Arguments:

            x(np.array): Array of shape `(n, m)` containing `n` inputs for which
                         to predict the conditional quantiles.

            use_splines(bool): Whether or not to use a spline fit to the CDF to
            approximate the PDF.

        Returns:

            Tuple (xs, fs) containing the :math: `x`-values in `xs` and corresponding
            values of the approximate posterior PDF :math: `F(x)` in `fs`.

        """

        y_pred = np.zeros(self.quantiles.size)
        y_pred = self.predict(x).ravel()

        y = np.zeros(y_pred.size + 1)
        y[1:-1] = 0.5 * (y_pred[1:] + y_pred[:-1])
        y[0] = 2 * y_pred[0] - y_pred[1]
        y[-1] = 2 * y_pred[-1] - y_pred[-2]

        if not use_splines:

            p = np.zeros(y.size)
            p[1:-1] = np.diff(self.quantiles) / np.diff(y_pred)
        else:

            y = np.zeros(y_pred.size + 2)
            y[1:-1] = y_pred
            y[0] = 3 * y_pred[0] - 2 * y_pred[1]
            y[-1] = 3 * y_pred[-1] - 2 * y_pred[-2]
            q = np.zeros(self.quantiles.size + 2)
            q[1:-1] = np.array(self.quantiles)
            q[0] = 0.0
            q[-1] = 1.0

            sr = CubicSpline(y, q, bc_type = "clamped")
            y = np.linspace(y[0], y[-1], 101)
            p = sr(y, nu = 1)

        return y, p


        y_pred = np.zeros(self.quantiles.size + 2)
        y_pred[1:-1] = self.predict(x)
        y_pred[0] = 2.0 * y_pred[1] - y_pred[2]
        y_pred[-1] = 2.0 * y_pred[-2] - y_pred[-3]

        if use_splines:
            x_t = np.zeros(x.size + 2)
            x_t[1:-1] = x
            x_t[0] = 2 * x[0] - x[1]
            x_t[-1] = 2 * x[-1] - x[-2]
            y_t = np.zeros(y.size + 2)
            y_t[1:-1] = y
            y_t[-1] = 1.0

        else:
            logger.info(y)
            x_new = np.zeros(x.size - 1)
            x_new[2:-2] = 0.5 * (x[2:-3] + x[3:-2])
            x_new[0:2] = x[0:2]
            x_new[-2:] = x[-2:]
            y_new = np.zeros(y.size - 1)
            y_new[1:-1] = np.diff(y[1:-1]) / np.diff(x[1:-1])
        return x_new, y_new

    def sample_posterior(self, x, n=1):
        r"""
        Generates :code:`n` samples from the estimated posterior
        distribution for the input vector :code:`x`. The sampling
        is performed by the inverse CDF method using the estimated
        CDF obtained from the :code:`cdf` member function.

        Arguments:

            x(np.array): Array of shape `(n, m)` containing `n` inputs for which
                         to predict the conditional quantiles.

            n(int): The number of samples to generate.

        Returns:

            Tuple (xs, fs) containing the :math: `x`-values in `xs` and corresponding
            values of the posterior CDF :math: `F(x)` in `fs`.
        """
        y_pred, qs = self.cdf(x)
        p = np.random.rand(n)
        y = np.interp(p, qs, y_pred)
        return y

    def posterior_mean(self, x):
        r"""
        Computes the posterior mean by computing the first moment of the
        estimated posterior CDF.

        Arguments:

            x(np.array): Array of shape `(n, m)` containing `n` inputs for which
                         to predict the posterior mean.
        Returns:

            Array containing the posterior means for the provided inputs.
        """
        y_pred, qs = self.cdf(x)
        mus = y_pred[-1] - np.trapz(qs, x=y_pred)
        return mus

    @staticmethod
    def crps(y_pred, y_test, quantiles):
        r"""
        Compute the Continuous Ranked Probability Score (CRPS) for given quantile
        predictions.

        This function uses a piece-wise linear fit to the approximate posterior
        CDF obtained from the predicted quantiles in :code:`y_pred` to
        approximate the continuous ranked probability score (CRPS):

        .. math::
            CRPS(\mathbf{y}, x) = \int_{-\infty}^\infty (F_{x | \mathbf{y}}(x')
            - \mathrm{1}_{x < x'})^2 \: dx'

        Arguments:

            y_pred(numpy.array): Array of shape `(n, k)` containing the `k`
                                 estimated quantiles for each of the `n`
                                 predictions.

            y_test(numpy.array): Array containing the `n` true values, i.e.
                                 samples of the true conditional distribution
                                 estimated by the QRNN.

            quantiles: 1D array containing the `k` quantile fractions :math:`\tau`
                       that correspond to the columns in `y_pred`.

        Returns:

            `n`-element array containing the CRPS values for each of the
            predictions in `y_pred`.
        """
        y_cdf = np.zeros((y_pred.shape[0], quantiles.size + 2))
        y_cdf[:, 1:-1] = y_pred
        y_cdf[:, 0] = 2.0 * y_pred[:, 1] - y_pred[:, 2]
        y_cdf[:, -1] = 2.0 * y_pred[:, -2] - y_pred[:, -3]

        ind = np.zeros(y_cdf.shape)
        ind[y_cdf > y_test.reshape(-1, 1)] = 1.0

        qs = np.zeros((1, quantiles.size + 2))
        qs[0, 1:-1] = quantiles
        qs[0, 0] = 0.0
        qs[0, -1] = 1.0

        return np.trapz((qs - ind)**2.0, y_cdf)

    def evaluate_crps(self, x, y_test):
        r"""
        Predict quantiles and compute the Continuous Ranked Probability Score (CRPS).

        This function evaluates the networks prediction on the
        inputs in `x` and evaluates the CRPS of the predictions
        against the materializations in `y_test`.

        Arguments:

            x(numpy.array): Array of shape `(n, m)` containing the `n`
                            `m`-dimensional inputs for which to evaluate
                            the CRPS.

            y_test(numpy.array): Array containing the `n` materializations of
                                 the true conditional distribution.

        Returns:

            `n`-element array containing the CRPS values for each of the
            inputs in `x`.

        """
        return QRNN.crps(self.predict(x), y_test, self.quantiles)

    def save(self, path):
        r"""
        Store the QRNN model in a file.

        This stores the model to a file using pickle for all
        attributes that support pickling. The Keras model
        is handled separately, since it can not be pickled.

        .. note:: In addition to the model file with the given filename,
                  additional files suffixed with :code:`_model_i` will be
                  created for each neural network this model consists of.

        Arguments:

            path(str): The path including filename indicating where to
                       store the model.

        """

        f = open(path, "wb")
        filename = os.path.basename(path)
        name = os.path.splitext(filename)[0]
        dirname = os.path.dirname(path)

        self.model_files = []
        for i, m in enumerate(self.models):
            self.model_files += [name + "_model_" + str(i)]
            m.save(os.path.join(dirname, self.model_files[i]))
        pickle.dump(self, f)
        f.close()

    @staticmethod
    def load(path):
        r"""
        Load a model from a file.

        This loads a model that has been stored using the `save` method.

        Arguments:

            path(str): The path from which to read the model.

        Return:

            The loaded QRNN object.
        """
        filename = os.path.basename(path)
        dirname = os.path.dirname(path)

        f = open(path, "rb")
        qrnn = pickle.load(f)
        qrnn.models = []
        for mf in qrnn.model_files:
            mf = os.path.basename(mf)
            try:
                mp = os.path.join(dirname, os.path.basename(mf))
                qrnn.models += [keras.models.load_model(mp, qrnn.custom_objects)]
            except:
                raise Exception("Error loading the neural network models. " \
                                "Please make sure all files created during the"\
                                " saving are in this folder.")
        f.close()
        return qrnn

    def __getstate__(self):
        dct = copy.copy(self.__dict__)
        dct.pop("models")
        return dct

    def __setstate__(self, state):
        self.__dict__ = state
        self.models = None



# -*- coding: utf-8 -*-
import numpy as np
from scipy.constants import h


def basic(LowerPop, UpperPop, Blu, Bul, Freq):
    u"""
    calculate absorption coefficient.\n
    Freq should be Hz
    """
    _const = h*Freq/4./np.pi
    return (LowerPop*Blu - UpperPop*Bul)*_const


def calc_abscoeff(tran, itera, alt=None):
    _up, _low = Tran_tag[tran][1:].astype(int)
    _v0 = Freqi[_up, _low]*1.e9
    _const = C.h*_v0/4./N.pi
    if alt is None:
        lowerPop = N.array(Ite_pop)[alt][itera][_low][0]
        upperPop = N.array(Ite_pop)[alt][itera][_up][0]
    else:
        lowerPop = N.array(Ite_pop)[:, itera, _low, 0]
        upperPop = N.array(Ite_pop)[:, itera, _up, 0]
    BLU = Blu[_up, _low]
    BUL = Bul[_up, _low]
    # AUL = Ai[_up, _low]
    return (lowerPop*BLU - upperPop*BUL)*_const



# -*- coding: utf-8 -*-
"""Functions to work with ASTER L1B satellite data.
"""
import datetime
import logging
import os
from collections import namedtuple

import gdal
import numpy as np
from scipy import interpolate
from skimage.measure import block_reduce

from typhon.math import multiple_logical


__all__ = [
    'ASTERimage',
    'lapserate_modis',
    'lapserate_moist_adiabate',
    'cloudtopheight_IR',
]

logger = logging.getLogger(__name__)

CONFIDENTLY_CLOUDY = 5
PROBABLY_CLOUDY = 4
PROBABLY_CLEAR = 3
CONFIDENTLY_CLEAR = 2


class ASTERimage:
    """ASTER L1B image object."""
    channels = (
        '1', '2', '3N', '3B',  # VNIR
        '4', '5', '6', '7', '8', '9',  # SWIR
        '10', '11', '12', '13', '14'  # TIR
    )

    def __init__(self, filename):
        """Initialize ASTER image object.

        Parameters:
            filename (str): Path to ASTER L1B HDF file.
        """
        self.filename = filename

        meta = self.get_metadata()
        SolarDirection = namedtuple('SolarDirection', ['azimuth', 'elevation']) # SolarDirection = (0< az <360, -90< el <90)
        self.solardirection = SolarDirection(
            *self._convert_metastr(meta['SOLARDIRECTION'], dtype=tuple)
        )
        self.sunzenith = 90 - self.solardirection.elevation
        
        self.datetime = datetime.datetime.strptime(
            meta['CALENDARDATE'] + meta['TIMEOFDAY'][:14], '%Y-%m-%d%H:%M:%S.%f0')
        SceneCenter = namedtuple('SceneCenter', ['latitude', 'longitude'])
        self.scenecenter = SceneCenter(
            *self._convert_metastr(meta['SCENECENTER'], dtype=tuple)
        )
        CornerCoordinates = namedtuple('CornerCoordinates', 
                                       ['LOWERLEFT', 'LOWERRIGHT', 
                                        'UPPERRIGHT', 'UPPERLEFT'])
        self.cornercoordinates = CornerCoordinates(
            self._convert_metastr(meta['LOWERLEFT'],dtype=tuple),
            self._convert_metastr(meta['LOWERRIGHT'], dtype=tuple),
            self._convert_metastr(meta['UPPERRIGHT'], dtype=tuple),
            self._convert_metastr(meta['UPPERLEFT'], dtype=tuple)
        )


    @property
    def basename(self):
        """Filename without path."""
        return os.path.basename(self.filename)

    @staticmethod
    def _convert_metastr(metastr, dtype=None):
        """Convert metadata data type."""
        if dtype is None:
            dtype = str

        if issubclass(dtype, tuple):
            return tuple(float(f.strip()) for f in metastr.split(','))
        else:
            return dtype(metastr)

    def get_metadata(self):
        """Read full ASTER metadata information."""
        return gdal.Open(self.filename).GetMetadata()

    def read_digitalnumbers(self, channel):
        """Read ASTER L1B raw digital numbers.

         Parameters:
             channel (str): ASTER channel number. '1', '2', '3N', '3B', '4',
                '5', '6', '7', '8', '9', '10', '11', '12', '13', '14'.

         Returns:
            ndarray: Digital numbers. Shape according to data resolution of
                the respective channel.
        """
        if channel in ('1', '2', '3N', '3B'):
            subsensor = 'VNIR'
        elif channel in ('4', '5', '6', '7', '8', '9'):
            subsensor = 'SWIR'
        elif channel in ('10', '11', '12', '13', '14'):
            subsensor = 'TIR'
        else:
            raise ValueError('The chosen channel is not supported.')

        data_path = (f'HDF4_EOS:EOS_SWATH:{self.filename}:{subsensor}_Swath:'
                     f'ImageData{channel}')
        swath = gdal.Open(data_path)

        data = swath.ReadAsArray().astype('float')
        data[data == 0] = np.nan  # Set edge pixels to NaN.

        return data

    def get_radiance(self, channel):
        """Get ASTER radiance values.

        Read digital numbers from ASTER L1B file and convert them to spectral
        radiance values in [W m-2 sr-1 um-1] at TOA by means of unit conversion
        coefficients ucc [W m-2 sr-1 um-1] and the gain settings at the sensor.

        See also:
            :func:`read_digitalnumbers`: Reads the raw digital numbers from
                ASTER L1B HDF4 files.

        Parameters:
             channel (str): ASTER channel number. '1', '2', '3N', '3B', '4',
                '5', '6', '7', '8', '9', '10', '11', '12', '13', '14'.

         Returns:
            ndarray: Radiance values at TOA. Shape according to data
                resolution of the respective channel.

        References:
            ﻿M. Abrams, S. H., & Ramachandran, B. (2002). Aster user handbook
            version 2, p.25 [Computer software manual]. Retrieved 2017-05-25,
            from https://asterweb.jpl.nasa.gov/content/03 data/04 Documents/
            aster user guide v2.pdf.
        """
        dn = self.read_digitalnumbers(channel)

        # Unit conversion coefficients ucc [W m-2 sr-1 um-1] for different gain
        # settings, high, normal, low1, and low2.
        ucc = {'1': (0.676, 1.688, 2.25, np.nan),
               '2': (0.708, 1.415, 1.89, np.nan),
               '3N': (0.423, 0.862, 1.15, np.nan),
               '3B': (0.423, 0.862, 1.15, np.nan),
               '4': (0.1087, 0.2174, 0.290, 0.290),
               '5': (0.0348, 0.0696, 0.0925, 0.409),
               '6': (0.0313, 0.0625, 0.0830, 0.390),
               '7': (0.0299, 0.0597, 0.0795, 0.332),
               '8': (0.0209, 0.0417, 0.0556, 0.245),
               '9': (0.0159, 0.0318, 0.0424, 0.265),
               '10': 6.882e-3,
               '11': 6.780e-3,
               '12': 6.590e-3,
               '13': 5.693e-3,
               '14': 5.225e-3,
               }

        if channel in ['1', '2', '3N', '3B', '4', '5', '6', '7', '8', '8',
                       '9']:
            meta = self.get_metadata()
            gain = meta[f'GAIN.{channel[0]}'].split(',')[1].strip()
            if gain == 'LO1':
                gain = 'LOW' # both refer to column 3 in ucc table.
            radiance = (dn - 1) * ucc[channel][
                                    ['HGH', 'NOR', 'LOW', 'LO2'].index(gain)]
        elif channel in ['10', '11', '12', '13', '14']:
            radiance = (dn - 1) * ucc[channel]
        else:
            raise ValueError('Invalid channel "{channel}".')

        return radiance

    def get_reflectance(self, channel):
        """Get ASTER L1B reflectance values at TOA.

        Spectral radiance values are converted to reflectance values for
        ASTER's near-infrared and short-wave channels (1-9).

        See also:
            :func:`get_radiance`: Reads the raw digital numbers from
                ASTER L1B HDF4 files and converts them to radiance values at
                TOA.

        Parameters:
             channel (str): ASTER channel number. '1', '2', '3N', '3B', '4',
                '5', '6', '7', '8', '9'.

        Returns:
            ndarray: Reflectance values at TOA. Shape according to data
                resolution of the respective channel.

        References:
        Thome, K.; Biggar, S.; Slater, P (2001). Effects of assumed solar
            spectral irradiance on intercomparisons of earth-observing sensors.
            In Sensors, Systems, and Next-Generation Satellites; Proceedings of
            SPIE; December 2001; Fujisada, H., Lurie, J., Weber, K., Eds.;
            Vol. 4540, pp. 260–269.
        http://www.pancroma.com/downloads/ASTER%20Temperature%20and%20Reflectance.pdf
        http://landsathandbook.gsfc.nasa.gov/data_prod/prog_sect11_3.html
        """
        # Mean solar exo-atmospheric irradiances [W m-2 um-1] at TOA according
        # to Thome et al. 2001.
        E_sun = (1848, 1549, 1114, 1114,
                 225.4, 86.63, 81.85, 74.85, 66.49, 59.85,
                 np.nan, np.nan, np.nan, np.nan, np.nan)

        sunzenith = 90 - self.solardirection.elevation

        doy = self.datetime.timetuple().tm_yday # day of the year (doy)

        distance = (0.016790956760352183
                    * np.sin(-0.017024566555637135 * doy
                             + 4.735251041365579)
                    + 0.9999651354429786) # acc. to Landsat Handbook

        return (np.pi * self.get_radiance(channel) * distance**2 /
                E_sun[self.channels.index(channel)] /
                np.cos(np.radians(sunzenith))
                )

    def get_brightnesstemperature(self, channel):
        """Get ASTER L1B reflectances values at TOA.

        Spectral radiance values are converted to brightness temperature values
        in [K] for ASTER's thermal channels (10-14).

        See also:
            :func:`get_radiance`: Reads the raw digital numbers from
                ASTER L1B HDF4 files and converts them to radiance values at
                TOA.

        Parameters:
             channel (str): ASTER channel number. '10', '11', '12', '13', '14'.

        Returns:
            ndarray: Brightness temperature values in [K] at TOA. Shape
                according to data resolution of the respective channel.

        References:
        http://www.pancroma.com/downloads/ASTER%20Temperature%20and%20Reflectan
            ce.pdf
        http://landsathandbook.gsfc.nasa.gov/data_prod/prog_sect11_3.html
        """
        K1 = {'10': 3040.136402, # Constant K1 [W m-2 um-1].
              '11': 2482.375199,
              '12': 1935.060183,
              '13': 866.468575,
              '14': 641.326517}

        K2 = {'10': 1735.337945, # Constant K2 [K].
              '11': 1666.398761,
              '12': 1585.420044,
              '13': 1350.069147,
              '14': 1271.221673}

        return (K2[channel] /
                np.log((K1[channel] / self.get_radiance(channel)) + 1)
                )

    def retrieve_cloudmask(self, output_binary=True, include_thermal_test=True,
                           include_channel_r5=True):
        """ASTER cloud mask.

        Four thresholding test based on visual bands distringuish between the
        dark ocean surface and bright clouds. An additional test corrects
        uncertain labeled pixels during broken cloud conditions and pixels with
        sun glint. A detailed description can be found in Werner et al., 2016.

        See also:
            :func:`get_reflectance`: Get reflectance values from ASTER's
                near-infrared and short-wave channels.
            :func:`get_brightnesstemperature`: Get brightness temperature
                values from ASTER's thermal channels.
            :func:multiple_logical`: Apply logical function to multiple
                arguments.

        Parameters:
            output_binary (bool): Switch between binary and full cloud mask
                flags.
                binary: 0 - clear (flag 2 & flag 3)
                        1 - cloudy (flag 4 & flag 5)
                full:   2 - confidently clear
                        3 - probably clear
                        4 - probably cloudy
                        5 - confidently cloudy
            include_thermal_test (bool): Switch for including test 5, which
                uses the thermal channel 14 at 11mu with 90m pixel resolution.
                The reduced resolution can introduce artificial straight cloud
                boundaries.
            include_channel_r5 (bool): Switch for including channel 5 in the
                thresholding tests. The SWIR sensor, including channel 5,
                suffered from temperature problems after May 2007. Per default,
                later recorded images are set to a value that has no influence
                on the thresholding tests.

        Returns:
            ndarray[float]: Cloud mask.

        References:
            Werner, F., Wind, G., Zhang, Z., Platnick, S., Di Girolamo, L.,
            Zhao, G., Amarasinghe, N., and Meyer, K.: Marine boundary layer
            cloud property retrievals from high-resolution ASTER observations:
            case studies and comparison with Terra MODIS, Atmos. Meas.
            Techannel., 9, 5869-5894, https://doi.org/10.5194/amt-9-5869-2016,
            2016.
        """

        # Read visual near infrared (VNIR) channels at 15m resolution.
        r1 = self.get_reflectance(channel='1')
        r2 = self.get_reflectance(channel='2')
        r3N = self.get_reflectance(channel='3N')

        # Read short-wave infrared (SWIR) channels at 30m resolution and match
        # VNIR resolution.
        r5 = self.get_reflectance(channel='5')
        if (self.datetime > datetime.datetime(2007, 5, 1)
            or not include_channel_r5):
            # The SWIR sensor suffered from temperature problems after May
            # 2007. Images later on are set to a dummy value "1", which won't
            # influence the following thresholding tests. Swath edge NaN pixels
            # stay NaN.
            r5[~np.isnan(r5)] = 1
        r5 = np.repeat(np.repeat(r5, 2, axis=0), 2, axis=1)

        # Read thermal (TIR) channel at 90m resolution and match VNIR
        # resolution.
        bt14 = self.get_brightnesstemperature(channel='14')
        bt14 = np.repeat(np.repeat(bt14, 6, axis=0), 6, axis=1)

        # Ratios for clear-cloudy-tests.
        r3N2 = r3N / r2
        r12 = r1 / r2

        ### TEST 1-4 ###
        # Set cloud mask to default "confidently clear".
        clmask = np.ones(r1.shape, dtype=np.float) * 2

        with np.warnings.catch_warnings():
            np.warnings.filterwarnings('ignore', r'invalid value encountered')

            # Set "probably clear" pixels.
            clmask[multiple_logical(r3N > 0.03, r5 > 0.01, 0.7 < r3N2,
                                    r3N2 < 1.75, r12 < 1.45,
                                    func=np.logical_and)] = PROBABLY_CLEAR


            # Set "probably cloudy" pixels.
            clmask[multiple_logical(r3N > 0.03, r5 > 0.015, 0.75 < r3N2,
                                    r3N2 < 1.75, r12 < 1.35,
                                    func=np.logical_and)] = PROBABLY_CLOUDY

            # Set "confidently cloudy" pixels
            clmask[multiple_logical(r3N > 0.065, r5 > 0.02, 0.8 < r3N2,
                                    r3N2 < 1.75, r12 < 1.2,
                                    func=np.logical_and)] = CONFIDENTLY_CLOUDY

            # Combine swath edge pixels.
            clmask[multiple_logical(np.isnan(r1), np.isnan(r2), np.isnan(r3N),
                                    np.isnan(r5), func=np.logical_or)] = np.nan

        if include_thermal_test:
            ### TEST 5 ###
            # Uncertain warm ocean pixels, higher than the 5th percentile of
            # brightness temperature values from all "confidently clear"
            # labeled pixels, are overwritten with "confidently clear".

            # Check for available "confidently clear" pixels.
            nc = np.sum(clmask == 2) / np.sum(~np.isnan(clmask))
            if (nc > 0.03):
                bt14_p05 = np.nanpercentile(bt14[clmask == 2], 5)
            else:
                # If less than 3% of pixels are "confidently clear", test 5
                # cannot be applied according to Werner et al., 2016. However,
                # a sensitivity study showed that combining "probably clear"
                # and "confidently clear" pixels in such cases leads to
                # plausible results and we derive a threshold correspondingly.
                bt14_p05 = np.nanpercentile(bt14[np.logical_or(clmask == 2,
                                                               clmask == 3)],5)

            with np.warnings.catch_warnings():
                np.warnings.filterwarnings('ignore',
                                           r'invalid value encountered')
                # Pixels with brightness temperature values above the 5th
                # percentile of clear ocean pixels are overwritten with
                # "confidently clear".
                clmask[np.logical_and(bt14 > bt14_p05, ~np.isnan(clmask))] = 2

            # Combine swath edge pixels.
            clmask[np.logical_or(np.isnan(clmask), np.isnan(bt14))] = np.nan

        if output_binary:
            clmask[np.logical_or(clmask == 2, clmask == 3)] = 0 # clear
            clmask[np.logical_or(clmask == 4, clmask == 5)] = 1 # cloudy

        return clmask

    def get_latlon_grid(self, channel='1'):
        """Create latitude-longitude-grid for specified channel data.

        A latitude-logitude grid is created from the corner coordinates of the
        ASTER image. The coordinates stated in the HDF4 meta data information
        correspond the edge pixels shifted by one in flight direction and to
        the left. The resolution and dimension of the image depends on the
        specified channel.

        Parameters:
            channel (str): ASTER channel number. '1', '2', '3N', '3B', '4',
                '5', '6', '7', '8', '9', '10', '11', '12', '13', '14'.

        Returns:
            ndarray[tuple]: Latitude longitude grid.

        References:
            M. Abrams, S. H., & Ramachandran, B. (2002). Aster user handbook
            version 2, p.25 [Computer software manual]. Retrieved 2017-05-25,
            from https://asterweb.jpl.nasa.gov/content/03 data/04 Documents/
            aster user guide v2.pdf.
        """
        meta = self.get_metadata()
        ImageDimension = namedtuple('ImageDimension', ['lon', 'lat'])
        imagedimension = ImageDimension(
            *self._convert_metastr(meta[f'IMAGEDATAINFORMATION{channel}'],
                                   tuple)[:2])

        # Image corner coordinates (file, NOT swath) according to the ASTER
        # Users Handbook p.57.
        LL = self._convert_metastr(meta['LOWERLEFT'],
                                 dtype=tuple)  # (latitude, longitude)
        LR = self._convert_metastr(meta['LOWERRIGHT'], dtype=tuple)
        UL = self._convert_metastr(meta['UPPERLEFT'], dtype=tuple)
        UR = self._convert_metastr(meta['UPPERRIGHT'], dtype=tuple)
        lat = np.array([[UL[0], UR[0]],
                        [LL[0], LR[0]]])
        lon = np.array([[UL[1], UR[1]],
                        [LL[1], LR[1]]])

        # Function for interpolation to full domain grid.
        flat = interpolate.interp2d([0, 1], [0, 1], lat, kind='linear')
        flon = interpolate.interp2d([0, 1], [0, 1], lon, kind='linear')

        # Grid matching shifted corner coordinates (see above).
        latgrid = np.linspace(0, 1, imagedimension.lat + 1)
        longrid = np.linspace(0, 1, imagedimension.lon + 1)

        # Apply lat-lon-function to grid and cut off last row and column to get
        # the non-shifted grid.
        lons = flon(longrid, latgrid)[:-1, :-1]
        lats = flat(longrid, latgrid)[:-1, :-1]

        return lats, lons

    def get_lapserate(self):
        """Estimate lapse rate from MODIS climatology using :func:`lapserate`."""
        return lapserate_modis(self.datetime.month, self.scenecenter[0])
    
    
    def get_cloudtopheight(self):
        """Estimate the cloud top height according to Baum et al., 2012 
        using :func: `cloudtopheight_IR`."""
        bt = self.get_brightnesstemperature('14')
        cloudmask = self.retrieve_cloudmask()
        latitude = self.scenecenter.latitude
        month = self.datetime.month
        
        return cloudtopheight_IR(bt, cloudmask, latitude, month, method='modis')
    
    
    def sensor_angles(self, sensor='vnir'):
        """Calculate a sun reflection angle for a given ASTER image depending on
        the sensor-sun geometry and the sensor settings.

        Note:
            All angular values are given in [°].

        Parameters:
            sensor (str): ASTER sensor ("vnir", "swir", or "tir").

        Returns:
            sensor_angle (ndarray): 2d field of size `cloudmask` of reflection
            angles for eachannel pixel.

        References:
             Kang Yang, Huaguo Zhang, Bin Fu, Gang Zheng, Weibing Guan, Aiqin Shi
             & Dongling Li (2015) Observation of submarine sand waves using ASTER
             stereo sun glitter imagery, International Journal of Remote Sensing,
             36:22, 5576-5592, DOI: 10.1080/01431161.2015.1101652
             Algorithm Theoretical Basis Document for ASTER Level-1 Data 
             Processing (Ver. 3.0).1996.
        """
        
        metadata = self.get_metadata()
        # Angular data from ASTER metadata data.
        S = float(metadata['MAPORIENTATIONANGLE'])
        
        if sensor=='vnir':
            P = float(metadata['POINTINGANGLE.1'])
            FOV = 6.09 # Field of view
            #IFOV = np.rad2deg(21.3e-6) Instantaneous field of view
            field = self.get_radiance(channel='1')
        elif sensor=='swir':
            P = float(metadata['POINTINGANGLE.2'])
            #IFOV = np.rad2deg(42.6e-6)
            FOV = 4.9
            field = self.get_radiance(channel='4')
        elif sensor=='tir':
            P = float(metadata['POINTINGANGLE.3'])
            #IFOV = np.rad2deg(127.8e-6)
            FOV = 4.9
            field = self.get_radiance(channel='10')
        else:
            raise ValueError('ASTER sensors are named: vnir, swir, tir.')

        # Instantaneous FOV (IVOF)
        #IFOV = FOV / cloudmask.shape[1]

        # Construct n-array indexing pixels n=- right and n=+ left from the image
        # center and perpendicular to flight direction.
        n = np.zeros(np.shape(field))
        nswath = []
        #assign indices n to pixel, neglegting the frist and last 5 swath rows
        for i in np.arange(start=5, stop=field.shape[0]-5, step=1, dtype=int):
            # for all scan rows: extract first and last not-NaN index 
            ind1 = (next(x[0] for x in enumerate(field[i]) if ~np.isnan(x[1]) ))
            #print('Index 1: ', ind1)
            ind2 = (field.shape[1] - next(x[0] for x in
                                    enumerate(field[i,5:-5][::-1]) if ~np.isnan(x[1]) ) )
            ind_mid = (ind1 + ind2) / 2
            nswath.append(ind2-ind1+1)
            # create indexing arrays for right and left side of swath
            right_arr = np.arange(start=-round(ind_mid), stop=0, step=1,
                                  dtype=int)*-1
            left_arr = np.arange(start=1,
                                 stop=field.shape[1] - round(ind_mid) + 1,
                                 step=1, dtype=int)*-1
            n[i] = np.asarray(list(right_arr) + list(left_arr))
            
        # add the first and last 5 scan lines as dupliates of the 6th and 6th-last row.
        n[:5] = n[5] # first
        n[-5:] = n[-6] # last

        n[np.isnan(field)] = np.nan
        
        # instantaneous field of view (IFOV)
        FOV_S = FOV / np.cos(np.deg2rad(S)) # correct for map orientation
        IFOV = FOV_S / np.median(nswath) # median scan line width

        # Thresholding index for separating left and right from nadir in
        # azimuth calculation.
        n_az = n * IFOV + P

        # ASTER zenith angle
        zenith = abs(np.asarray(n) * IFOV + P)

        # ASTER azimuth angle
        azimuth = np.ones(field.shape) * np.nan
        # swath right side
        azimuth[np.logical_and(n_az<0, ~np.isnan(n))] = 90 + S
        # swath left side
        azimuth[np.logical_and(n_az>=0, ~np.isnan(n))] = 270 + S

        return (zenith, azimuth)


    def reflection_angles(self):
        """Calculate the reflected sun angle, theta_r, of specular reflection
        of sunlight into an instrument sensor.

        Returns:
            (ndarray): 2d field of size `cloudmask` of reflection
            angles in [°] for eachannel pixel.

        References:
             Kang Yang, Huaguo Zhang, Bin Fu, Gang Zheng, Weibing Guan, Aiqin Shi
             & Dongling Li (2015) Observation of submarine sand waves using ASTER
             stereo sun glitter imagery, International Journal of Remote Sensing,
             36:22, 5576-5592, DOI: 10.1080/01431161.2015.1101652
        """
        sun_azimuth = self.solardirection.azimuth + 180 # +180° corresponds to "anti-/reflected" sun
        sun_zenith = 90 - self.solardirection.elevation # sun zenith = 90 - sun elevation
        
        sensor_zenith, sensor_azimuth = self.sensor_angles()

        return np.degrees(np.arccos(np.cos(np.deg2rad(sensor_zenith))
                                    * np.cos(np.deg2rad(sun_zenith))
                                    + np.sin(np.deg2rad(sensor_zenith))
                                    * np.sin(np.deg2rad(sun_zenith))
                                    * np.cos( np.deg2rad(sensor_azimuth)
                                    - np.deg2rad(sun_azimuth)) ) )


def lapserate_modis(month, latitude):
    """Estimate of the apparent lapse rate in [K/km].
    Typical lapse rates are assumed for each month and depending on the
    location on Earth, i.e. southern hemisphere, tropics, or northern
    hemisphere. For a specific case the lapse rate is estimated by a 4th order
    polynomial and polynomial coefficients given in the look-up table.
    This approach is based on the MODIS cloud top height retrieval and applies
    to data recorded at 11 microns. According to Baum et al., 2012, the
    predicted lapse rates are restricted to a maximum and minimum of
    10 and 2 K/km, respectively.


    Parameters:
        month (int): Month of the year.
        latitude (float): Latitude of center coordinate.

    Returns:
        float: Lapse rate estimate.

    References:
        Baum, B.A., W.P. Menzel, R.A. Frey, D.C. Tobin, R.E. Holz, S.A.
        Ackerman, A.K. Heidinger, and P. Yang, 2012: MODIS Cloud-Top Property
        Refinements for Collection 6. J. Appl. Meteor. Climatol., 51,
        1145–1163, https://doi.org/10.1175/JAMC-D-11-0203.1
    """
    lapserate_lut = {
        'month': np.arange(1, 13),
        'SH_transition': np.array([-3.8, -21.5, -2.8, -23.4, -12.3, -7.0,
                                   -10.5, -7.8, -8.6, -7.0, -9.2, -3.7]),
        'NH_transition': np.array([22.1, 12.8, 10.7, 29.4, 14.9, 16.8,
                                   15.0, 19.5, 17.4, 27.0, 22.0, 19.0]),
        'a0': np.array([(2.9769801, 2.9426577, 1.9009563),  # Jan
                        (3.3483239, 2.6499606, 2.4878736),  # Feb
                        (2.4060296, 2.3652047, 3.1251275),  # Mar
                        (2.6522387, 2.5433158, 13.3931707),  # Apr
                        (1.9578263, 2.4994028, 1.6432070),  # May
                        (2.7659754, 2.7641496, -5.2366360),  # Jun
                        (2.1106812, 3.1202043, -4.7396481),  # Jul
                        (3.0982174, 3.4331195, -1.4424843),  # Aug
                        (3.0760552, 3.4539390, -3.7140186),  # Sep
                        (3.6377215, 3.6013337, 8.2237401),  # Oct
                        (3.3206165, 3.1947419, -0.4502047),  # Nov
                        (3.0526633, 3.1276377, 9.3930897)]),  # Dec
        'a1': np.array([(-0.0515871, -0.0510674, 0.0236905),
                        (0.1372575, -0.0105152, -0.0076514),
                        (0.0372002, 0.0141129, -0.1214572),
                        (0.0325729, -0.0046876, -1.2206948),
                        (-0.2112029, -0.0364706, 0.1151207),
                        (-0.1186501, -0.0728625, 1.0105575),
                        (-0.3073666, -0.1002375, 0.9625734),
                        (-0.1629588, -0.1021766, 0.4769307),
                        (-0.2043463, -0.1158262, 0.6720954),
                        (-0.0857784, -0.0775800, -0.5127533),
                        (-0.1411094, -0.1045316, 0.2629680),
                        (-0.1121522, -0.0707628, -0.8836682)]),
        'a2': np.array([(0.0027409, 0.0052420, 0.0086504),
                        (0.0133259, 0.0042896, 0.0079444),
                        (0.0096473, 0.0059242, 0.0146488),
                        (0.0100893, 0.0059325, 0.0560381),
                        (-0.0057944, 0.0082002, 0.0033131),
                        (0.0011627, 0.0088878, -0.0355440),
                        (-0.0090862, 0.0064054, -0.0355847),
                        (-0.0020384, 0.0010499, -0.0139027),
                        (-0.0053970, 0.0015450, -0.0210550),
                        (0.0024313, 0.0041940, 0.0205285),
                        (-0.0026068, 0.0049986, -0.0018419),
                        (-0.0009913, 0.0055533, 0.0460453)]),
        'a3': np.array([(0.0001136, 0.0001097, -0.0002167),
                        (0.0003043, 0.0000720, -0.0001774),
                        (0.0002334, -0.0000159, -0.0003188),
                        (0.0002601, 0.0000144, -0.0009874),
                        (-0.0001050, 0.0000844, -0.0001458),
                        (0.0000937, 0.0001768, 0.0005188),
                        (-0.0000890, 0.0002620, 0.0005522),
                        (0.0000286, 0.0001616, 0.0001759),
                        (-0.0000541, 0.00017117, 0.0002974),
                        (0.0001495, 0.0000941, -0.0003016),
                        (0.0000058, 0.0001911, -0.0000369),
                        (0.0000180, 0.0001550, -0.0008450)]),
        'a4': np.array([(0.00000113, -0.00000372, 0.00000151),
                        (0.00000219, -0.0000067, 0.00000115),
                        (0.00000165, -0.00000266, 0.00000210),
                        (0.00000199, -0.00000346, 0.00000598),
                        (-0.00000074, -0.00000769, 0.00000129),
                        (0.00000101, -0.00001168, -0.00000262),
                        (0.00000004, -0.00001079, -0.00000300),
                        (0.00000060, 0.00000510, -0.00000080),
                        (-0.00000002, 0.00000248, -0.00000150),
                        (0.00000171, -0.0000041, 0.00000158),
                        (0.00000042, -0.00000506, 0.00000048),
                        (0.00000027, -0.00000571, 0.00000518)]),
    }

    ind_month = month - 1

    if latitude < lapserate_lut['SH_transition'][ind_month]:
        region_flag = 0  # Southern hemisphere
    elif (latitude >= lapserate_lut['SH_transition'][ind_month] and
          latitude <= lapserate_lut['NH_transition'][ind_month]):
        region_flag = 1  # Tropics
    elif latitude > lapserate_lut['NH_transition'][ind_month]:
        region_flag = 2  # Northern hemisphere
    else:
        raise ValueError('Latitude of center coordinate cannot be read.')

    lapserate = (
            lapserate_lut['a0'][ind_month][region_flag]
            + lapserate_lut['a1'][ind_month][region_flag] * latitude
            + lapserate_lut['a2'][ind_month][region_flag] * latitude ** 2
            + lapserate_lut['a3'][ind_month][region_flag] * latitude ** 3
            + lapserate_lut['a4'][ind_month][region_flag] * latitude ** 4
    )

    return lapserate


def lapserate_moist_adiabate():
    """Moist adiabatic lapse rate in [K/km].
    """
    return 6.5


def cloudtopheight_IR(bt, cloudmask, latitude, month, method='modis'):
    """Cloud Top Height (CTH) from 11 micron channel.
    
    Brightness temperatures (bt) are converted to CTHs using the IR window approach: 
    (bt_clear - bt_cloudy) / lapse_rate.

    See also:
        :func:`skimage.measure.block_reduce`
            Down-sample image by applying function to local blocks.
        :func:`lapserate_moist_adiabate`
            Constant value 6.5 [K/km]
        :func:`lapserate_modis`
            Estimate of the apparent lapse rate in [K/km]
            depending on month and latitude acc. to Baum et al., 2012.

    Parameters:
        bt (ndarray): brightness temperatures form 11 micron channel.
        cloudmask (ndarray): binary cloud mask.
        month (int): month of the year.
        latitude (ndarray): latitudes in [°], positive North, negative South.
        method (str): approach used to derive CTH: 'modis' see Baum et al., 2012,
            'simple' uses the moist adiabatic lapse rate.

    Returns:
        ndarray: cloud top height.
    
    References:
        Baum, B.A., W.P. Menzel, R.A. Frey, D.C. Tobin, R.E. Holz, S.A.
        Ackerman, A.K. Heidinger, and P. Yang, 2012: MODIS Cloud-Top Property
        Refinements for Collection 6. J. Appl. Meteor. Climatol., 51,
        1145–1163, https://doi.org/10.1175/JAMC-D-11-0203.1
    """
    # Lapse rate
    if method=='simple':
        lapserate = lapserate_moist_adiabate()

    elif method=='modis':
        lapserate = lapserate_modis(month, latitude)
    else:
        raise ValueError("Method is not supported.")

    resolution_ratio = np.shape(cloudmask)[0] // np.shape(bt)[0]

    cloudmask_inverted = cloudmask.copy()
    cloudmask_inverted[np.isnan(cloudmask_inverted)] = 1
    cloudmask_inverted = np.asarray(np.invert(np.asarray(
                                cloudmask_inverted, dtype=bool)), dtype=int)

    cloudmask[np.isnan(cloudmask)] = 0
    cloudmask = np.asarray(cloudmask, dtype=int)

    # Match resolutions of cloud mask and brightness temperature (bt) arrays.
    if resolution_ratio > 1:
        # On bt resolution, flag pixels as cloudy only if all subgrid pixels
        # are cloudy in the original cloud mask.
        mask_cloudy = block_reduce(cloudmask,
                                        (resolution_ratio, resolution_ratio),
                                        func=np.alltrue)
        # Search for only clear pixels to derive a bt clearsky/ocean value.
        mask_clear = block_reduce(cloudmask_inverted,
                                        (resolution_ratio, resolution_ratio),
                                        func=np.alltrue)
    elif resolution_ratio < 1:
        try:
            mask_cloudy = np.repeat(np.repeat(cloudmask, resolution_ratio,
                                        axis=0), resolution_ratio, axis=1)
            mask_clear = np.repeat(np.repeat(cloudmask_inverted,
                                                  resolution_ratio, axis=0),
                                                  resolution_ratio, axis=1)
        except ValueError:
            raise ValueError(
                'Problems matching the shapes of provided cloud mask and bt arrays.')
    else:
        mask_cloudy = cloudmask.copy()
        mask_clear = cloudmask_inverted.copy()

    bt_cloudy = np.ones(np.shape(bt)) * np.nan
    bt_cloudy[mask_cloudy] = bt[mask_cloudy]
    bt_clear_avg = np.nanmean(bt[mask_clear])

    return (bt_clear_avg - bt_cloudy) / lapserate



# -*- coding: utf-8 -*-
"""Testing the functions in typhon.physics.
"""
import numpy as np

from typhon import physics


class TestEM:
    """Testing the typhon.physics.em functions."""
    def test_planck(self):
        """Test calculation of Planck black body radiation."""
        x = physics.planck(183e9, 273.15)
        assert np.allclose(x, 2.7655049656659619e-15)

    def test_planck_wavelength(self):
        """Test calculation of Planck black body radiation (for wavelength)."""
        x = physics.planck_wavelength(0.01, 273.15)
        assert np.allclose(x, 0.00022552294907196693)

    def test_planck_wavenumber(self):
        """Test calculation of Planck black body radiation (for wavenumber)."""
        x = physics.planck_wavenumber(100, 273.15)
        assert np.allclose(x, 2.2552294907196695e-08)

    def test_rayleighjeans(self):
        """Test calculation Rayleigh-Jeans black body radiation."""
        x = physics.rayleighjeans(183e9, 273.15)
        assert np.allclose(x, 2.810445098218606e-15)

    def test_rayleighjeans_wavelength(self):
        """Test calculation RJ black body radiation (for wavelength)."""
        x = physics.rayleighjeans_wavelength(0.01, 273.15)
        assert np.allclose(x, 0.00022611794774492813)

    def test_radiance2planckTb(self):
        """Test conversion between spectral radiance and Planck Tb."""
        f = np.linspace(100, 80e12, 100)
        r = physics.planck(f, 300)

        tb = physics.radiance2planckTb(f, r)

        assert np.allclose(tb, 300)

    def test_radianc2rayleighjeansTb(self):
        """Test conversion between spectral radiance and Rayleigh-Jeans Tb."""
        f = np.linspace(100, 80e12, 100)
        r = physics.rayleighjeans(f, 300)

        tb = physics.radiance2rayleighjeansTb(f, r)

        assert np.allclose(tb, 300)

    def test_snell(self):
        """Test Snell's law."""
        x = physics.snell(1, 1, 60)
        assert np.allclose(x, 60)

    def test_fresnel(self):
        """Test Fresnel law."""
        x = physics.fresnel(0.6, 1, 45)
        ref = (0.13098189212919023, -0.3619142054813409)
        assert np.allclose(x, ref)

    def test_frequency2wavelength(self):
        """Test conversion from frequency to wavelength."""
        x = physics.frequency2wavelength(183e9)
        assert np.allclose(x, 0.0016382101530054644)

    def test_frequency2wavenumber(self):
        """Test conversion from frequency to wavenumber."""
        x = physics.frequency2wavenumber(183e9)
        assert np.allclose(x, 610.4222942126182)

    def test_wavelength2frequency(self):
        """Test conversion from wavelength to frequency."""
        x = physics.wavelength2frequency(0.01)
        assert np.allclose(x, 29979245800.0)

    def test_wavelength2wavenumber(self):
        """Test conversion from wavelength to wavenumber."""
        x = physics.wavelength2wavenumber(0.01)
        assert x == 100

    def test_wavenumber2frequency(self):
        """Test conversion from wavenumber to frequency."""
        x = physics.wavelength2frequency(100)
        assert np.allclose(x, 2997924.58)

    def test_wavenumber2wavelength(self):
        """Test conversion from wavenumber to wavelength."""
        x = physics.wavelength2wavenumber(100)
        assert x == 0.01


class TestThermodynamics:
    """Testing the typhon.physics.thermodynamics functions."""
    def test_e_eq_ice_mk(self):
        """Test calculation of equilibrium water vapor pressure over ice."""
        x = physics.e_eq_ice_mk(260)
        assert np.allclose(x, 195.81934571953161)

    def test_e_eq_water_mk(self):
        """Test calculation of equilibrium water vapor pressure over water."""
        x = physics.e_eq_water_mk(280)
        assert np.allclose(x, 991.85662101784112)

    def test_e_eq_mixed_mk(self):
        """Test calculation of vapor pressure with respect to mixed phase."""
        assert np.allclose(physics.e_eq_ice_mk(240),
                           physics.e_eq_mixed_mk(240))

        assert np.allclose(physics.e_eq_water_mk(290),
                           physics.e_eq_mixed_mk(290))

        assert (physics.e_eq_ice_mk(263)
                < physics.e_eq_mixed_mk(263)
                < physics.e_eq_water_mk(263))

    def test_density(self):
        """Test calculation of air density."""
        x = physics.density(1013.15e2, 273.15)
        assert np.allclose(x, 1.2921250376634072)

    def test_mixing_ratio2specific_humidity(self):
        """Test conversion of mass mixing ratio to specific humidity."""
        q = physics.mixing_ratio2specific_humidity(0.02)
        assert np.isclose(q, 0.0196078431372549)

    def test_mixing_ratio2vmr(self):
        """Test conversion of mass mixing ratio to VMR."""
        x = physics.mixing_ratio2vmr(0.02)
        assert np.isclose(x, 0.03115371853180794)

    def test_specific_humidity2mixing_ratio(self):
        """Test conversion of specific humidity to mass mixing ratio."""
        w = physics.specific_humidity2mixing_ratio(0.02)
        assert np.isclose(w, 0.020408163265306124)

    def test_specific_humidity2vmr(self):
        """Test conversion of specific humidity to VMR."""
        x = physics.specific_humidity2vmr(0.02)
        assert np.isclose(x, 0.03176931009073226)

    def test_vmr2mixing_ratio(self):
        """Test conversion of VMR to mass mixing ratio."""
        w = physics.vmr2mixing_ratio(0.04)
        assert np.isclose(w, 0.025915747437955664)

    def test_vmr2specific_humidity(self):
        """Test conversion of VMR to specific humidity."""
        q = physics.vmr2specific_humidity(0.04)
        assert np.isclose(q, 0.025261087474946833)



from datetime import datetime
import warnings

import numpy as np
import xarray as xr

from .common import HDF4, expects_file_info

pyhdf_is_installed = False
try:
    from pyhdf import HDF, VS, V
    from pyhdf.SD import SD, SDC
    pyhdf_is_installed = True
except ImportError:
    pass

__all__ = [
    'CloudSat',
]


class CloudSat(HDF4):
    """File handler for CloudSat data in HDF4 files.
    """

    # This file handler always wants to return at least time, lat and lon
    # fields. These fields are required for this:
    standard_fields = {
        "UTC_start",
        "Profile_time",
        "Latitude",
        "Longitude"
    }

    # Map the standard fields to standard names:
    mapping = {
        "Latitude": "lat",
        "Longitude": "lon",
        "dim_0": "scnline",
    }

    def __init__(self, **kwargs):

        # Call the base class initializer
        super().__init__(**kwargs)

    @expects_file_info()
    def get_info(self, file_info, **kwargs):
        """Return a :class:`FileInfo` object with parameters about the
        file content.

        Args:
            file_info: Path and name of the file of which to retrieve the info
                about.
            **kwargs: Additional keyword arguments.

        Returns:
            A FileInfo object.
        """

        file = SD(file_info.path, SDC.READ)
        file_info.times[0] = \
            datetime.strptime(getattr(file, 'start_time'), "%Y%m%d%H%M%S")
        file_info.times[1] = \
            datetime.strptime(getattr(file, 'end_time'), "%Y%m%d%H%M%S")

        return file_info

    @expects_file_info()
    def read(self, file_info, **kwargs):
        """Read and parse HDF4 files and load them to a xarray.Dataset

        A description about all variables in CloudSat dataset can be found in
        http://www.cloudsat.cira.colostate.edu/data-products/level-2c/2c-ice?term=53.

        Args:
            file_info: Path and name of the file as string or FileInfo object.
            **kwargs: Additional keyword arguments that are valid for
                :class:`typhon.files.handlers.common.HDF4`.

        Returns:
            A xarray.Dataset object.
        """

        # We need to import at least the standard fields
        user_fields = kwargs.pop("fields", {})
        fields = self.standard_fields | set(user_fields)

        # We catch the user mapping here, since we do not want to deal with
        # user-defined names in the further processing. Instead, we use our own
        # mapping
        user_mapping = kwargs.pop("mapping", None)

        # Load the dataset from the file:
        dataset = super().read(
            file_info, fields=fields, mapping=self.mapping, **kwargs
        )

        dataset["time"] = self._get_time_field(dataset, file_info)

        # Remove fields that we do not need any longer (expect the user asked
        # for them explicitly)
        dataset = dataset.drop(
            {"UTC_start", "Profile_time"} - set(user_fields),
        )

        if user_mapping is not None:
            dataset = dataset.rename(user_mapping)

        return dataset

    def _get_time_field(self, dataset, file_info):
        # This gives us the starting time of the first profile in seconds
        # since midnight in UTC:
        first_profile_time = round(dataset['UTC_start'].item(0))

        # This gives us the starting time of all other profiles in seconds
        # since the start of the first profile.
        profile_times = dataset['Profile_time']

        # Convert the seconds to milliseconds
        profile_times *= 1000
        profile_times = profile_times.astype("int")

        try:
            date = file_info.times[0].date()
        except AttributeError:
            # We have to load the info by ourselves:
            date = self.get_info(file_info).times[0].date()

        # Put all times together so we obtain one full timestamp
        # (date + time) for each data point. We are using the
        # starting date coming from parsing the filename.
        profile_times = \
            np.datetime64(date) \
            + np.timedelta64(first_profile_time, "s") \
            + profile_times.astype("timedelta64[ms]")

        return profile_times



"""Collection of classes related to filtering

"""

# Any commits made to this module between 2015-05-01 and 2019-02-01
# by Gerrit Holl are developed for the EC project “Fidelity and
# Uncertainty in Climate Data Records from Earth Observations (FIDUCEO)”.
# Grant agreement: 638822
# 
# All those contributions are dual-licensed under the MIT license for use
# in typhon, and the GNU General Public License version 3.

import sys
import abc
import dbm
import logging
import tempfile
import pathlib
import shutil
import datetime
import warnings

import numpy
try:
    import progressbar
except ImportError:
    progressbar = None

from . import dataset

logger = logging.getLogger(__name__)

class FilterError(Exception):
    """For any errors related to filtering.
    """

class OutlierFilter(metaclass=abc.ABCMeta):
    
    @abc.abstractmethod
    def filter_outliers(self, C):
        ...

class MEDMAD(OutlierFilter):
    """Outlier filter based on Median Absolute Deviation

    """

    def __init__(self, cutoff, fallback_min_std=0.1):
        self.cutoff = cutoff
        self.fallback_min_std = 0.1
    
    def filter_outliers(self, C):
        cutoff = self.cutoff
        if C.ndim == 3:
            med = numpy.ma.median(
                C.reshape(C.shape[0]*C.shape[1], C.shape[2]),
                0)
            mad = numpy.ma.median(
                abs(C - med).reshape(C.shape[0]*C.shape[1], C.shape[2]),
                0)
            if (mad==0).any():
                # use fallback
                med[mad==0] = C[..., mad==0].reshape(C.shape[0]*C.shape[1], (mad==0).sum()).mean(0)
                mad[mad==0] = numpy.c_[
                    C[..., mad==0].reshape(C.shape[0]*C.shape[1], (mad==0).sum()).std(0),
                    numpy.tile(self.fallback_min_std, (mad==0).sum())].max(1)
        elif C.ndim < 3:
            med = numpy.ma.median(C.reshape((-1,)))
            mad = numpy.ma.median(abs(C - med).reshape((-1,)))
            if mad==0:
                med = C.mean()
                mad = max(C.std(), self.fallback_min_std)
        else:
            raise ValueError("Cannot filter outliers on "
                "input with ndim={ndim:d}>3 dimensions".format(ndim=C.ndim))
        fracdev = ((C - med)/mad)
        return abs(fracdev) > cutoff

class OrbitFilter:
    """Generic, abstract class for any kind of filtering.

    Implementations of this class are intended to be used between and
    after each orbit file is read by Dataset.read_period.  One important
    implementation for this class are the various OverlapFilter
    implementations, that make sure overlaps are removed.  This can either
    be done after each orbit or when all orbits have been read.

    Different filters may need different meta-information that needs to be
    provided by specific Dataset._read implementations.  For example, the
    FirstlineDBFilter needs a header name (or the entire header from which
    to extract this).  Each filterer can state what keyword arguments need
    to be passed on to the reading routine using the args_to_reader
    attribute.

    Moments in filtering that this class takes is responsible for:

    - Before the first orbit is read: .reset()
    - After each orbit is read. .filter(...)
    - After the last orbit is read: .finalise(...)

    You can also use this class for validation, i.e. to check that data
    are correct and if they aren't, raise an exception, if they are,
    return data as-is.
    """

    args_to_reader = {}

    def reset(self):
        pass

    def filter(self, scanlines, **extra):
        return scanlines

    def finalise(self, arr):
        return arr
    
    def __eq__(self, other):
        return self.__dict__ == other.__dict__

class TimeMaskFilter(OrbitFilter):
    """Throw out bad (masked) times.
    """

    def __init__(self, ds):
        self.ds = ds

    def reset(self):
        pass

    def filter(self, scanlines, **extra):
        # when time is masked, we should REALLY despair.  We want
        # to have sequential scanlines.  Throw them out already!
        if scanlines["time"].mask.any():
            logger.warning("Throwing out {:d} scanlines because "
                "their times are flagged and not sequential".format(
                    scanlines["time"].mask.sum()))
            good = ~scanlines["time"].mask
            return scanlines[good] 
        else:
            return scanlines

    def finalise(self, arr):
        return arr

class HIRSTimeSequenceDuplicateFilter(OrbitFilter):
    """Force scanlines to be in the proper sequence without duplicates

    This is for handling time sequence and duplicate issues within a
    single orbit file.  The two have to be in one filter because the
    application of the latter relies on the application of the first.
    I suppose one could conceivably want to remove time sequence problems
    but keep duplicate scanlines.
    """

    def reset(self):
        pass

    def filter(self, scanlines, **extra):
        goodorder = scanlines["hrs_scnlin"][1:] > scanlines["hrs_scnlin"][:-1]
        if not goodorder.all():
            logger.warning("{:d} scanlines are out of "
                "order, resorting".format(
                (~goodorder).sum()))
            neworder = numpy.argsort(scanlines["hrs_scnlin"].data)
            scanlines = scanlines[neworder]        

        # if there still are any now, it can only be due to duplicates
        goodorder = scanlines["hrs_scnlin"][1:] > scanlines["hrs_scnlin"][:-1]
        if not goodorder.all():
            logger.warning("{:d} duplicate "
                "scanlines (judging from scanline number), removing".format((~goodorder).sum()))
            (_, ii) = numpy.unique(scanlines["hrs_scnlin"],
                                   return_index=True)
            scanlines = scanlines[ii]


        # still time sequence issues?
        goodtime = numpy.argsort(scanlines["time"]) == numpy.arange(scanlines.size)
        if not goodtime.all():
            logger.warning("Still has time sequence issues! "
                "Dropping {:d} scanlines to be on the safe side. "
                "This is probably overconservative.".format((~goodtime).sum()))
            scanlines = scanlines[goodtime]

        # in some cases, like 1985-11-30T17:19:45.056 on NOAA-9,
        # there are scanlines with different scanline numbers but
        # the same time!
        (_, ii) = numpy.unique(scanlines["time"], return_index=True)
        if ii.size < scanlines["time"].size:
            logger.warning("Oops!  There are scanlines with different "
                "scanline numbers but the same time!  Removing {:d} "
                "more lines.  I hope that's it!".format(
                    scanlines["time"].size-ii.size))
            scanlines = scanlines[ii]

        return scanlines

    def finalise(self, arr):
        return arr

class HIRSFlagger(OrbitFilter):
    """Apply HIRS flags and raise error in case of failure
    """

    def __init__(self, ds, max_flagged=0.5):
        self.ds = ds
        self.max_flagged = max_flagged

    def reset(self):
        pass

    def filter(self, scanlines, **extra):
        scanlines = self.ds.get_mask_from_flags(extra["header"],
            scanlines, max_flagged=self.max_flagged)

        return scanlines

    def finalise(self, arr):
        return arr

class HIRSCalibCountFilter(OrbitFilter):
    """Apply masking based on calibration count filter for HIRS
    """

    def __init__(self, ds, filter_calibcounts:OutlierFilter):
        self.ds = ds
        self.filter_calibcounts = filter_calibcounts

    def reset(self):
        pass

    def filter(self, scanlines, **extra):

        for v in self.ds.views:
            x = scanlines[self.ds.scantype_fieldname] == getattr(self.ds,
                    "typ_{:s}".format(v))
            if not x.any():
                raise dataset.InvalidDataError("Out of {:d} scanlines, "
                    "found no {:s} views, cannot calibrate!".format(
                        scanlines.shape[0], v))
            scanlines.mask["counts"][x, 8:, :] = self.ds.filter_calibcounts.filter_outliers(
                scanlines["counts"][x, 8:, :])


        cc = scanlines["calcof_sorted"]
#        scanlines = self.ds.apply_calibcount_filter(scanlines)
        if cc.ndim == 4:
            calibzero = (cc[:, :, 1, :]==0).all(2)
        if cc.ndim == 3:
            calibzero = (cc==0).all(2)
        if "bt" in scanlines.dtype.names:
            scanlines["bt"].mask[...] |= calibzero[:, numpy.newaxis, :19]
        if "radiance" in scanlines.dtype.names:
            scanlines["radiance"].mask[...] |= calibzero[:, numpy.newaxis, :20]
        try:
            # if one is masked, so should the other…
            scanlines["radiance"].mask[:, :, :19] |= scanlines["bt"].mask
        except ValueError: # xarray raises ValueError, not KeyError
            # not a problem if either of those fields does not exists
            pass

        return scanlines

    def finalise(self, arr):
        return arr

class HIRSPRTTempFilter(OrbitFilter):
    """Apply masking to PRT counts when they are outliers
    """

    def __init__(self, ds, filter_prttemps:OutlierFilter):
        self.ds = ds
        self.filter_prttemps = filter_prttemps

    def reset(self):
        pass

    def filter(self, scanlines, **extra):
        for fld in [nm for nm in scanlines.dtype.names if
                    nm.startswith("temp_")]:
            scanlines[fld].mask |= self.ds.filter_prttemps.filter_outliers(
                                    scanlines[fld])
        return scanlines

    def finalise(self, arr):
        return arr

class OverlapFilter(OrbitFilter):
    """Implementations to feed into firstline filtering

    This is used in tovs HIRS reading routine.
    """
    
    # the 'late' attribute should be set to True if all duplicate checking
    # is done late; this will suppress after-orbit verification of absense
    # of duplicates.
    late = False

class FirstlineDBFilter(OverlapFilter):
    def __init__(self, ds, granules_firstline_file):
        self.ds = ds
        self.granules_firstline_file = granules_firstline_file

    def reset(self):
        pass

    _tmpdir = None
    _firstline_db = None
    def filter(self, scanlines, header):
        """Filter out any scanlines that existed in the previous granule.

        Only works on datasets implementing get_dataname from the header.
        """
        dataname = self.ds.get_dataname(header, robust=True)
        if self._firstline_db is None:
            try:
                self._firstline_db = dbm.open(
                    str(self.granules_firstline_file), "r")
            except dbm.error as e: # presumably a lock
                tmpdir = tempfile.TemporaryDirectory()
                self._tmpdir = tmpdir # should be deleted only when object is
                tmp_gfl = str(pathlib.Path(tmpdir.name,
                    self.granules_firstline_file.name))
                logger.warning("Cannot read GFL DB at {!s}: {!s}, "
                    "presumably in use, copying to {!s}".format(
                        self.granules_firstline_file, e.args, tmp_gfl))
                shutil.copyfile(str(self.granules_firstline_file),
                    tmp_gfl)
                self.granules_firstline_file = tmp_gfl
                self._firstline_db = dbm.open(tmp_gfl)
        try:
            firstline = int(self._firstline_db[dataname])
        except KeyError as e:
            raise FilterError("Unable to filter firstline: {:s}".format(
                e.args[0])) from e
        if firstline > scanlines.shape[0]:
            logger.warning("Full granule {:s} appears contained in previous one. "
                "Refusing to return any lines.".format(dataname))
            return scanlines[0:0]
        return scanlines[scanlines["hrs_scnlin"] > firstline]    

    def update_firstline_db(self, satname=None, start_date=None, end_date=None,
            overwrite=False):
        """Create / update the firstline database

        Create or update the database describing for each granule what the
        first scanline is that doesn't occur in the preceding granule.

        If a granule is entirely contained within the previous one,
        firstline is set to L+1 where L is the number of lines.
        """
        prev_head = prev_line = None
        satname = satname or self.ds.satname
        start_date = start_date or self.ds.start_date
        end_date = end_date or self.ds.end_date
        if end_date > datetime.datetime.now():
            end_date = datetime.datetime.now()
        logger.info("Updating firstline-db {:s} for "
            "{:%Y-%m-%d}--{:%Y-%m-%d}".format(satname, start_date, end_date))
        count_updated = count_all = 0
        with dbm.open(str(self.granules_firstline_file), "c") as gfd:
            try:
                bar = progressbar.ProgressBar(max_value=1,
                    widgets=[progressbar.Bar("=", "[", "]"), " ",
                        progressbar.Percentage(), ' (',
                        progressbar.AdaptiveETA(), " -> ",
                        progressbar.AbsoluteETA(), ') '])
            except AttributeError:
                dobar = False
                bar = None
                logger.info("If you had the "
                    "progressbar2 module, you would have gotten a "
                    "nice progressbar.")
            else:
                dobar = sys.stdout.isatty()
                if dobar:
                    bar.start()
                    bar.update(0)
            for (g_start, gran) in self.ds.find_granules_sorted(start_date, end_date,
                            return_time=True, satname=satname):
                try:
                    (cur_line, extra) = self.ds.read(gran,
                        apply_scale_factors=False, calibrate=False)
                    cur_head = extra["header"]
                    cur_time = self.ds._get_time(cur_line)
                except (dataset.InvalidFileError,
                        dataset.InvalidDataError) as exc:
                    logger.error("Could not read {!s}: {!s}".format(gran, exc))
                    continue
                lab = self.ds.get_dataname(cur_head, robust=True)
                if lab in gfd and not overwrite:
                    logger.debug("Already present: {:s}".format(lab))
                elif prev_line is not None:
                    # what if prev_line is None?  We don't want to define any
                    # value for the very first granule we process, as we might
                    # be starting to process in the middle...
                    if cur_time.max() > prev_time.max():
                        # Bugfix 2017-01-16: do not get confused between
                        # the index and the hrs_scnlin field.  So far, I'm using
                        # the index to set firstline but the hrs_scnlin
                        # field to apply it.
                        #first = (cur_time > prev_time[-1]).nonzero()[0][0]
                        # Bugfix 2017-08-21: instead of taking the last
                        # time from the previous granule, take the
                        # maximum; this allows for time sequence errors.
                        # See #139
                        first = cur_line["hrs_scnlin"][cur_time > prev_time.max()].min()
                        logger.debug("{:s}: {:d}".format(lab, first))
                    else:
                        first = cur_line["hrs_scnlin"].max()+1
                        logger.info("{:s}: Fully contained in {:s}!".format(
                            lab, self.ds.get_dataname(prev_head, robust=True)))
                    gfd[lab] = str(first)
                    count_updated += 1
                prev_line = cur_line.copy()
                prev_head = cur_head.copy()
                prev_time = cur_time.copy()
                if dobar:
                    bar.update((g_start-start_date)/(end_date-start_date))
                count_all += 1
            if dobar:
                bar.update(1)
                bar.finish()
            logger.info("Updated {:d}/{:d} granules".format(count_updated, count_all))

    def finalise(self, arr):
        return arr

class NullLineFilter(OverlapFilter):
    """Do not filter firstlines at all
    """

    def reset(self):
        pass

    def filter(self, scanlines, header):
        return scanlines

    def finalise(self, arr):
        return arr

class HIRSBestLineFilter(OverlapFilter):
    """Choose best between overlaps.

    Currently works only for HIRS.
    """

    # those are expected to be different between orbits, others are not
    knowndiff = {'hrs_scnlin', 'hrs_qualind', 'hrs_linqualflgs',
                 'hrs_chqualflg'}
    warn_overlap = False # set to True if you want warnings when overlaps inconsistent

    late = True
    def __init__(self, ds, warn_overlap=False):
        self.ds = ds
        self.warn_overlap = warn_overlap

    orbits = []
    def reset(self):
        self.orbits.clear()

    def filter(self, scanlines, header):
        self.orbits.append(
            (self.ds.get_dataname(header, robust=True),
             scanlines["time"].min(),
             scanlines["time"].max(),
             scanlines.size))
        return scanlines

    def select_winner(self, rep):
        """Select "winning" scanline among 2 or more

        Between 2 or more identical scanlines, select the one that fits
        best.  That probably means least flags etc.

        Takes a ndarray with dtype containing at least the HIRS flag
        fields.  Dimension should be (n,) where n>1.

        Returns the index of the best choice.
        """
        
        # relevant fields:
        #
        # - hrs_qualind
        # - hrs_linqualflgs
        # - hrs_chqualflg
        # - hrs_mnfrqual
        #
        # But this differs between HIRS/2/3/4...
        # need some neutral way of "scoring" how bad it is

        scores = self.ds.flagscore(rep)
        return numpy.argmin(scores)

    def finalise(self, arr, verify_overlap_consistency=False):
        """For all sets of duplicates, select optimal one

        Arguments:

            arr [ndarray], dtype with structured ndarray according to what
            the HIRS reading routine produces

            verify_overlap_consistency [bool].  Verify that overlapping
            scanlines have consistent information.  Defaults to False
            because this can be very slow.

        Returns:

            same array but sorted and with duplicates removed; for each
            pair of duplicates, best alternative is chosen
        """
        # Find all pairs of duplicates (usually in sets of 2) and the
        # with corresponding indices and multiplicity (count).  
        arrsrt = arr[numpy.argsort(arr["time"])]
        _, ii, cnt = numpy.unique(arrsrt["time"],
            return_index=True, return_counts=True)
        logger.debug("Selecting optimal scanlines for "
            f"{ii.size:d} overlapping pairs")
        # (mult_ii, mult_cnt) take the same values as if I were to do:
        # for (mult_ii, mult_cnt) in zip(ii[cnt>1], cnt[cnt>1]), but I
        # want to keep the indices so I can /write/ to ii
        multcnt_i_all = (cnt>1).nonzero()[0]
        for multcnt_i in multcnt_i_all:
            mult_ii = ii[multcnt_i]
            mult_cnt = cnt[multcnt_i]
            rep = arrsrt[mult_ii:mult_ii+mult_cnt]
            if self.warn_overlap:
                fields_notclose = {nm for nm in rep.dtype.names
                    if not
                    (rep[nm][0]==rep[nm]
                     if rep[nm].dtype.kind[0] in "MmS"
                     else numpy.isclose(rep[nm][0, ...], rep[nm])
                    ).all()} - self.knowndiff
                if len(fields_notclose) > 0:
                    warnings.warn(
                        "Overlapping or duplicate scanlines "
                        "have inconsistent values for ".format(
                            rep[0]["time"].astype(datetime.datetime))
                        + ", ".join(list(fields_notclose)),
                            UserWarning)
            # ii normally contains the index of the first of a sequence of
            # duplicates; select_winner returns the index of the optimal
            # choice within the set 'rep' of repeated scanlines; the sum
            # will thus be the index of our scanline of choice
            ii[multcnt_i] += self.select_winner(rep)
        return arrsrt[ii]
